'''
Data translators

Some :class:`~owmeta_core.datasource.DataSource` and `~owmeta_core.datasource.DataTranslator`
types. Some deal with generic file types (e.g., comma-separated values) while others are
specific to the format of a kind of file housed in owmeta.
'''
